#include <mpi.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>

//tagy pro specifikaci typu zpravy
//predavane z farmeru do workeru
#define TAG_WORK 1
#define TAG_STOP 2
#define TAG_FARM 3

#define FARMER_PROCESS_ID 0

#define MAX_LINE_SIZE 200


//struktura s parametry workeru
typedef struct
{
	int dim;
	int wrk_pt;
	int id;

}WORKER_INFO; 

//struktura s parametry farmeru
typedef struct
{
	// so - pole se sloupci matice - nemeni se
	// re - radky matice - aktualni vysledek
	// tm - pomocne pole - pro ukladani mezivypoctu

	int *so,*re,*tm;

	int elm_co;
	int dim;

	int *cntrl;
	int wrks_co;
	int wrk_pt;
	int exponent;

}FARMER_INFO; 

//vypise matici na standardni vystup
void print_matrix(int* elements, int count, unsigned char align)
{
	int i;
	
	int dim = (int) sqrt(count);

	for (i = 0; i < count; i++)
	{
		printf(" %5d ", elements[i]);
		
		if ( ((i + 1) % dim) == 0 && align && i != 0)
		{
			printf("\n");
		}
	}

	printf("\n");
}

//provadi prevod matice z radkove na sloupcovou formu
int* get_cols_from_row_matrix(int *rows, int dimension, int count)
{
	int* cols = (int *) malloc(count*sizeof(int));
	
	int i, k, j = 0;
					
	for (i = 0; i < dimension; i++)
	{
		for (k = i; k < count; k += dimension)
		{
			cols[j++] = rows[k];
		}
	}
	
	return cols;
}


//vytvari buffer pro worker - radky a sloupce jsou razeny za sebou po dvojicich
//poradi se urcuje z leveho horniho rohu matice - staci jedna souradnice begin, count urcuje pocet prvku
int* get_work_part_buffer(int* multiplicant, int* multiplier, int dimension, int begin, int count)
{
	int k = 0;
	int i = 0;
	
	int* buffer = (int *) calloc(count * dimension * 2, sizeof(int));
	
	int b_col = begin % dimension;
	int b_row = begin / dimension;

	int buffer_index = 0;
	
	for (i = 0; i < count; i++)
	{
		for (k = 0; k < dimension; k++)
		{
			buffer[k + buffer_index] = multiplicant[(b_row * dimension) + k];
			buffer[k + buffer_index + dimension] = multiplier[(b_col * dimension) + k];
		}
		
		b_col++;

		if (b_col == dimension)
		{
			b_col = 0; b_row++;
		}
						
		buffer_index += 2 * dimension;
	}

	return buffer;
}

//posila inicializacni cast prace workerum 
//vraci index dalsiho prvku pro zpracovani
int send_init_parts_to_workers(FARMER_INFO *f_inf)
{
	int i = 0;
	int work_index = 0;
	int process_count = 0;

	int work_part = f_inf->wrk_pt;

	//pro zpracovani staci jeden proces !!! jednotka objemu prace je vetsi nez pocet vsech 
	//elementu matice
	if (f_inf->wrk_pt > f_inf->elm_co)
	{
		work_part = f_inf->elm_co;
	}
	
	//zjistime zda pri danem objemu prace na jeden proces budou potreba 
	//vsechny procesy
	
	//maximalni pocet procesu, tak aby kazdy dostal jeden pracovni kus 
	int maximal_process_count = f_inf->elm_co / work_part;

	if ((f_inf->elm_co % f_inf->wrk_pt) > 0)
	{
		maximal_process_count++;
	}

	//potrebny pocet procesu je vetsi nez zadany
	if (maximal_process_count > f_inf->wrks_co)
	{
		process_count = f_inf->wrks_co;
	}
	
	//potrebujeme mene procesu nez bylo zadano - predimenzovano, neuplatni se load balancing
	if (f_inf->wrks_co > maximal_process_count)
	{
		process_count = maximal_process_count;
	}

	for (i = 0; i < process_count; i++)
	{
		int *send_buffer = get_work_part_buffer(f_inf->re, f_inf->so, f_inf->dim, work_index, work_part);

		MPI_Send(send_buffer, work_part * f_inf->dim * 2 , MPI_INT, i + 1, TAG_WORK, MPI_COMM_WORLD);
		
		free(send_buffer);

		f_inf->cntrl[i] = work_index;

		work_index += work_part;

		if ((work_index + work_part) > f_inf->elm_co)
		{
			work_part = f_inf->elm_co - work_index;
		}
	}

	return work_index;
}

//ceka na prijeti vysledku od workeru
//vraci pocet prijatych elementu
int receive_part_from_worker(FARMER_INFO *f_inf, int *proc_id)
{
	MPI_Status stats;
	int work_part = 0;
	int i = 0;

	MPI_Probe(MPI_ANY_SOURCE, TAG_FARM, MPI_COMM_WORLD, &stats);
			
	MPI_Get_count(&stats, MPI_INT, &work_part);
	
	int *recv_buffer = (int *) calloc(work_part, sizeof(int));
	
	MPI_Recv(recv_buffer, work_part, MPI_INT, MPI_ANY_SOURCE, TAG_FARM, MPI_COMM_WORLD, &stats);

	*proc_id = stats.MPI_SOURCE - 1;

	int result_index = f_inf->cntrl[*proc_id];

	for (i = 0; i < work_part; i++)
	{
		f_inf->tm[result_index + i] = recv_buffer[i];
	}
	
	free(recv_buffer);

	return work_part;
}

//funkce rozdelujici praci workerum
void farmer(FARMER_INFO *f_inf)
{
	int k = 0;
	int i = 0;
	int j = 0;
		
	int work_index = 0;
	int elm_co_ready = 0;
	int work_part = f_inf->wrk_pt;  
	int proc_id = 0;

	for (k = 0; k < f_inf->exponent; k++ )
	{
		work_index = send_init_parts_to_workers(f_inf);	

		elm_co_ready = 0;

		do
		{

			elm_co_ready += receive_part_from_worker(f_inf, &proc_id);
		
			if (work_index != f_inf->elm_co)
			{
				if ((work_index + work_part) > f_inf->elm_co)
				{
					work_part = f_inf->elm_co - work_index;
				}	 
				else
				{
					work_part = f_inf->wrk_pt;
				}
			
				int *send_buffer = get_work_part_buffer(f_inf->re, f_inf->so, f_inf->dim, work_index, work_part);
   		
				MPI_Send(send_buffer, work_part * f_inf->dim * 2 , MPI_INT, proc_id + 1, TAG_WORK, MPI_COMM_WORLD);

				f_inf->cntrl[proc_id] = work_index;

				free(send_buffer);

				work_index += work_part;
			}

		}while (elm_co_ready < f_inf->elm_co);
		
		memcpy(f_inf->re, f_inf->tm, f_inf->elm_co * sizeof(int));
	}
}

//zpracovava pozadavky od farmare a posila zpet vysledky
void worker(WORKER_INFO *w_inf)
{
	MPI_Status stats;
	
	int* recv_buffer; 
	int* send_buffer; 

	int i = 0;
	int j = 0;
	int work_part = 0;
	int count = 0;

	while(1)
	{
		MPI_Probe(FARMER_PROCESS_ID, MPI_ANY_TAG, MPI_COMM_WORLD, &stats);
			
		MPI_Get_count(&stats, MPI_INT, &count);
    	
		recv_buffer = (int *) calloc(count, sizeof(int));

		MPI_Recv(recv_buffer, count, MPI_INT, FARMER_PROCESS_ID, MPI_ANY_TAG, MPI_COMM_WORLD, &stats);
		
		//!!! STOP worker
		if (stats.MPI_TAG == TAG_STOP)
		{
			free(recv_buffer);

			break;
		}
		
		work_part = count / (2 * w_inf->dim);

		send_buffer = (int *) calloc(work_part, sizeof(int));

		for (i = 0; i < work_part; i++)
		{
			for (j = 0; j < w_inf->dim; j++)
			{
				//v prijmacim bufferu jsou radky a sloupce ulozeny za sebou
				send_buffer[i] += recv_buffer[(i * w_inf->dim * 2) + j] * recv_buffer[(i * w_inf->dim * 2) + w_inf->dim + j];
			}
		}
		
		MPI_Send(send_buffer, work_part, MPI_INT, FARMER_PROCESS_ID, TAG_FARM, MPI_COMM_WORLD);

		free(send_buffer);
		free(recv_buffer);
	}
}

//funkce pro nacteni matice ze souboru
int* read_matrix_from_file(char *file_name, int* dimension, int *element_count)
{
	FILE *fr;

	
	char line[MAX_LINE_SIZE];
	int i = 0, elm_index = 0;
	int *mx;

	fr = fopen(file_name,"r");

	if (fr == NULL)
	{
		return NULL;
	}

	fgets(line, MAX_LINE_SIZE - 1, fr);

	*dimension = atoi(line);

	if (dimension <= 0)
	{
		return NULL;
	}
	
	*element_count = *dimension * *dimension;

	mx = (int *) calloc(*element_count, sizeof(int));

	while (fgets(line, MAX_LINE_SIZE - 1, fr) != NULL)
	{
		for (i = 0; i < MAX_LINE_SIZE -1; i++)
		{
			if (isdigit(line[i]) && (*element_count > elm_index))
			{
				mx[elm_index++] = atoi(line + i);
			}

			while(isdigit(line[i]) && i < (MAX_LINE_SIZE - 1))
			{
				i++;
			}
		}
	}

	return mx;

}

//vygenerovani nahodne matice
int *generate_random_matrix(int element_count)
{
	int *mx = (int *) calloc(element_count, sizeof(int));
	int i = 0;

	for (i = 0; i < element_count; i++)
	{
		mx[i] = (int)(rand() / (((double)RAND_MAX + 1)/ 10));
	}

	return mx;
}

//nacteni matice ze vstupu
int *read_from_input(int* dimension, int* work_part, int* exponent, int* elm_co)
{
	int i, k, *mx;
		
	printf("Rad matice = ");
	scanf("%d", dimension);
	
	printf("Exponent = ");
	scanf("%d", exponent);
	
	*elm_co = (*dimension) * (*dimension);

	mx = (int *) calloc(*elm_co, sizeof(int));

	printf("Pocet prvku (skalarnich soucinu) na jeden proces = ");
	scanf("%d", work_part);

	for (i = 0; i < *dimension; i++)
	{
		for (k = 0; k < *dimension; k++)
		{
			printf("[%d, %d] = ", i, k);
			scanf("%d", mx + (i * (*dimension) + k));
		}
	}

	putchar('\n');

	return mx;
}

//zastaveni vsech workeru
void stop_workers(int proc_num)
{
	int send_stop = TAG_STOP;
	int i;

	for (i = 0; i < proc_num - 1; i++)
	{
		MPI_Send(&send_stop, 1, MPI_INT, i + 1, TAG_STOP, MPI_COMM_WORLD);
	}
}

//vraci defaultni info v pripade ze byly zadany neplatne argumenty pri spousteni programu
FARMER_INFO* default_farmer_info(int proc_num)
{
	int i = 0;
	
	FARMER_INFO* f_inf = (FARMER_INFO*) calloc(1, sizeof(FARMER_INFO));

	f_inf->wrks_co = proc_num - 1;

	f_inf->cntrl = (int *) calloc(proc_num - 1, sizeof(int));

	f_inf->wrk_pt = 2;
	f_inf->dim = 3;
	f_inf->elm_co = 9;
	f_inf->exponent = 2;

	f_inf->re = (int *) calloc(f_inf->elm_co, sizeof(int));
				
	for (i = 0; i < f_inf->elm_co; i++)
	{
		f_inf->re[i] = i;
	}

	f_inf->so = get_cols_from_row_matrix(f_inf->re, f_inf->dim, f_inf->elm_co);
	f_inf->tm = (int *) calloc(f_inf->elm_co, sizeof(int));

	return f_inf;
}

//vytvari souhrn parametru pro farmare - predavano ve strukture pro lepsi strukturovani programu
FARMER_INFO* init_farmer_info(int argc, char* argv[], int proc_num)
{
	int i;
	
	if (argc == 1)
	{
		return NULL;
	}

	FARMER_INFO* f_inf = (FARMER_INFO*) calloc(1, sizeof(FARMER_INFO));

	f_inf->wrks_co = proc_num - 1;

	f_inf->cntrl = (int *) calloc(proc_num - 1, sizeof(int));
	
	if (strcmp(argv[1],"-i") == 0)
	{
		f_inf->re = read_from_input(&(f_inf->dim), &(f_inf->wrk_pt), &(f_inf->exponent), &(f_inf->elm_co));
	}
	
	if (strcmp(argv[1],"-r") == 0)
	{
		f_inf->dim = atoi(argv[2]);
		f_inf->exponent = atoi(argv[3]);
		f_inf->wrk_pt = atoi(argv[4]);

		f_inf->elm_co = f_inf->dim * f_inf->dim;

		f_inf->re = generate_random_matrix(f_inf->elm_co);
	}

	if (strcmp(argv[1],"-f") == 0)
	{
		f_inf->exponent = atoi(argv[2]);
		f_inf->wrk_pt = atoi(argv[3]);

		f_inf->re = read_matrix_from_file(argv[4],&(f_inf->dim), &(f_inf->elm_co));
	}
	
	if (f_inf->re == NULL)
	{
		free(f_inf->cntrl);
		free(f_inf);
		
		return NULL;
	}

	f_inf->so = get_cols_from_row_matrix(f_inf->re, f_inf->dim, f_inf->elm_co);
	f_inf->tm = (int *) calloc(f_inf->elm_co, sizeof(int));

	return f_inf;
}

//zasle zakladni parametry ulohy vsem workerum
void send_worker_info(int wrk_pt, int dim)
{
	int send_buffer[2];

	send_buffer[0] = wrk_pt;
	send_buffer[1] = dim;
	
	MPI_Bcast(send_buffer, 2, MPI_INT, FARMER_PROCESS_ID, MPI_COMM_WORLD); 
	
}

//inicializuje strukturu s parametry pro worker
WORKER_INFO* init_worker_info(int my_id)
{
	MPI_Status stats;

	int recv_buffer[2];
	
	WORKER_INFO* w_inf = (WORKER_INFO*) calloc(1, sizeof(WORKER_INFO));

	MPI_Bcast(recv_buffer, 2, MPI_INT, FARMER_PROCESS_ID, MPI_COMM_WORLD); 

	w_inf->wrk_pt = recv_buffer[0];
	w_inf->dim	 = recv_buffer[1];
	w_inf->id = my_id;

	return w_inf;
}

//uvolni strukturu s parametry farmeru
void free_farmer_info(FARMER_INFO* f_inf)
{
	free(f_inf->re);
	free(f_inf->so);
	free(f_inf->tm);
	free(f_inf);
}

//vypise parametry na std_out
void print_par_info()
{
	printf("\nParametry pro spusteni :");
	printf("\n\nNahodna matice : ");
	printf("\n\t\t -r <dimenze> <exponent> <objem prace>");
	printf("\n\nVstup ze souboru : ");
	printf("\n\t\t -f <exponent> <objem prace> <nazev souboru>");
	printf("\n\nVstup z konzole : ");
	printf("\n\t\t -i");
}

//vstupni bod programu - nacteni parametru, spusteni farmeru a workeru
int main(int argc, char *argv[])
{
	int num_procs;
	int my_id;

	MPI_Init(&argc,&argv);

	MPI_Comm_size(MPI_COMM_WORLD,&num_procs);
	MPI_Comm_rank(MPI_COMM_WORLD,&my_id);
	
	setbuf(stdout,NULL);

	if (my_id == FARMER_PROCESS_ID)
	{
		
		FARMER_INFO *f_inf = init_farmer_info(argc, argv, num_procs);
	
		

		if (f_inf == NULL)
		{
			print_par_info();

			printf("\n\n !!! Spatne vstupni argumenty - poustim defaultni ulohu !!! \n\n");
			f_inf = default_farmer_info(num_procs);

			printf("Exponent = %d", f_inf->exponent);
		}

		send_worker_info(f_inf->wrk_pt, f_inf->dim);

		printf("\nZdroj:\n");
		print_matrix(f_inf->re, f_inf->elm_co, 1);

		printf("\nExponent = %d\n", f_inf->exponent);

		long time_stamp = time(NULL);

		if (f_inf->exponent == 0)
		{
			int i;

			for ( i = 0; i < f_inf->elm_co; i++)
			{
				f_inf->re[i] = 1;
			}
		}
		else
		{
			f_inf->exponent = f_inf->exponent - 1;
			farmer(f_inf);
		}

		stop_workers(num_procs);
			
		printf("Vysledek :\n");
		print_matrix(f_inf->re, f_inf->elm_co, 1);

		free_farmer_info(f_inf);

		printf("Doba trvani vypoctu %ld s \n", time(NULL) - time_stamp);
	}
	else
	{
		WORKER_INFO *w_inf = init_worker_info(my_id);
		
		worker(w_inf);

		free(w_inf);
	}

	MPI_Finalize(); 

	return 0;
}
